#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>

#include <lq_com.h>
#include <lq_nl.h>
#include <lq_q.h>
#include <lq_c.h>
#include <lq_f.h>


/* External variables */


/* Internal variables */
short	lq_debug = 0;
short	lq_inited = 0;
char	lq_strerror0[1024];

/* Functions */
char *lq_strerror(void)
{
	return lq_strerror0;
}

int lq_getstats(short type, int ifindex, void (*f) (lq_stats *)) {
	int	ret;

	ret = lq_nl_getstats(type, ifindex, f);
	snprintf (lq_strerror0, sizeof(lq_strerror0),
		"%s", lq_nl_strerror());

	return ret;
}

char *lq_dump(lq *O)
{
	static char	ret[4096];
	char		tmp[128];
	char		*p;

	snprintf(ret, sizeof (ret), "Dump object @ %p)...\n", (void *) O);

	if (!O) {
		snprintf (tmp, sizeof (tmp), "NULL object!\n");
		strcat(ret, tmp);
		goto out;
	}

	snprintf(tmp, sizeof(tmp), "Type = %s [%d]",
		lq_type (O), O->type);
	strcat(ret, tmp);

	snprintf(tmp, sizeof(tmp), ", subtype = %s [%d]",
		lq_subtype (O), O->subtype);
	strcat(ret, tmp);

	if (O->parent != TC_H_ROOT)
		snprintf(tmp, sizeof(tmp), ", parent 0x%04x:0x%04x",
			O->parent >> 16, O->parent & 0xffff);
	else
		snprintf(tmp, sizeof(tmp), ", root");
	strcat(ret, tmp);

	snprintf(tmp, sizeof(tmp), ", handler 0x%04x:0x%04x",
		O->id >> 16, O->id & 0xffff);
	strcat(ret, tmp);

	snprintf(tmp, sizeof(tmp), ", ifindex %d\n",
		O->ifindex);
	strcat(ret, tmp);

	/* dump stats */
	p = lq_dump_stats(O);
	strcat(ret, p);
	strcat(ret, "\n");

	/* dump rest */
	switch (O->type) {
		case LQ_qdisc: 	p = lq_q_dump(O);	break;
		case LQ_class: 	p = lq_c_dump(O);	break;
		case LQ_filter:	p = lq_f_dump(O);	break;
		default:		p = "Unknown type";
	}
	strcat(ret, p);

	out:
	return ret;
}

void lq_stop(void)
{
	if (lq_inited == 0)
		return;

	lq_nl_stop();

	lq_inited = 0;
}

int lq_set(lq *O)
{
	int		ret = 0;
	unsigned int	type, subtype;

	if (!O) {
		snprintf (lq_strerror0, sizeof(lq_strerror0),
			"set: object is null");
		ret = -1;
		goto out;
	}

	type = O->type;
	subtype = O->subtype;

	/* zero */
	memset ((void *) O, 0, sizeof (*O));

	O->magic = 0xadadadad;

	O->type = type;
	O->subtype = subtype;

	O->parent = TC_H_ROOT;

	/* set default */
	switch (O->type) {
		case LQ_qdisc:	lq_q_set(O);		break;
		case LQ_class:	lq_c_set(O);		break;
		case LQ_filter:	lq_f_set(O);		break;
		default:
			snprintf(lq_strerror0, sizeof(lq_strerror0),
				"set: invalid type");
			ret = -1;
			goto out;
	}

	out:
	return ret;
}

int lq_init(void)
{
	int	ret;

	if (lq_inited == 1) {
		ret = 0;
		goto out;
	}

	lq_nl_debug = lq_debug;

	ret = lq_nl_init();
	if (ret == -1) {
		snprintf(lq_strerror0, sizeof (lq_strerror0),
			"init: cannot init nl [%s]", lq_nl_strerror ());
		ret = -1;
		goto out;
	}

	ret = lq_read_psched();
	if (ret == -1) {
		snprintf (lq_strerror0, sizeof (lq_strerror0),
			"init: cannot read/parse /proc/net/psched");
		ret = -1;
		goto out;
	}

	lq_inited = 1;
	atexit(lq_stop);
	ret = 0;

	out:
	return ret;
}

int lq_do(short cmd, lq *O)
{
	int			ret;
	struct lq_nl_pak	req;

	if (lq_inited == 0) {
		snprintf(lq_strerror0, sizeof(lq_strerror0),
			"do: call init first");
		ret = -1;
		goto out;
	}

	if (!O) {
		snprintf(lq_strerror0, sizeof(lq_strerror0),
			"do: object is NULL");
		ret = -1;
		goto out;
	}

	if (O->magic != 0xadadadad) {
		snprintf(lq_strerror0, sizeof(lq_strerror0),
			"do: please call lq_set on this object first");
		ret = -1;
		goto out;
	}

	/* encode personal data */
	ret = lq_nl_encode (cmd, O, &req);
	if (ret == -1) {
		snprintf (lq_strerror0, sizeof (lq_strerror0),
			"do: cannot encode specific data [%s]",
			lq_nl_strerror ());
		ret = -1;
		goto out;
	}

	ret = lq_nl_send (&req);
	if (ret == -1) {
		snprintf (lq_strerror0, sizeof (lq_strerror0),
			"do: cannot send data with nl [%s]",
			lq_nl_strerror ());
		ret = -1;
		goto out;
	}

	ret = 0;

	out:
	return ret;
}

int lq_getif (char *dev)
{
	int	ret;

	if (lq_inited == 0) {
		snprintf (lq_strerror0, sizeof (lq_strerror0),
			"lq_init was not called first!");
		ret = -1;
		goto out;
		}

	ret = lq_nl_getif (dev);
	snprintf (lq_strerror0, sizeof (lq_strerror0), "%s", lq_nl_strerror ());

	out:
	return ret;
}

void lq_listifs (void)
{
	lq_nl_listifs ();
}
