#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>

#include <lq_com.h>


/* Variables */
static __u32	lq_t2us = 1;
static __u32	lq_us2t = 1;
static double	lq_tick_in_usec = 1;


/* Functions */
char *lq_q_type(lq *O)
{
	char *ret;

	switch (O->subtype) {
		case LQ_q_htb:		ret = "htb";	break;
		case LQ_q_sfq:		ret = "sfq";	break;
		case LQ_q_pfifo:	ret = "pfifo";	break;
		case LQ_q_bfifo:	ret = "bfifo";	break;
		case LQ_q_prio:		ret = "prio";	break;
		case LQ_q_dup:		ret = "dup";	break;
		case LQ_q_ooo:		ret = "ooo";	break;
		case LQ_q_esfq:		ret = "esfq";	break;
		default:		ret = "unknown";
	}

	return ret;
}

char *lq_c_type(lq *O)
{
	char    *ret;

	switch (O->subtype) {
		case LQ_c_htb:	ret = "htb";	break;
		default:	ret = "unknown";
	}

	return ret;
}

char *lq_f_type(lq *O)
{
	char    *ret;

	switch (O->subtype) {
		case LQ_f_u32:		ret = "u32";		break;
		case LQ_f_fw:		ret = "fw";		break;
		case LQ_f_route:	ret = "route";		break;
		case LQ_f_tcindex:	ret = "tcindex";	break;
		default:		ret = "unknown";
	}

	return ret;
}

char *lq_subtype(lq *O)
{
	char	*ret;

	switch (O->type) {
		case LQ_qdisc:
			ret = lq_q_type(O);
			break;
		case LQ_class:
			ret = lq_c_type(O);
			break;
		case LQ_filter:
			ret = lq_f_type(O);
			break;
		default:
			ret = "unknown";
	}

	return ret;
}

char *lq_type(lq *O)
{
	char	*ret;

	switch (O->type) {
		case LQ_qdisc: 	ret = "qdisc";	break;
		case LQ_class: 	ret = "class";	break;
		case LQ_filter:	ret = "filter";	break;
		default:		ret = "unknown";
	}

	return ret;
}

char *lq_dump_stats(lq *O)
{
	static char	ret [128];

	snprintf(ret, sizeof(ret), "bytes=%llu packets=%u drops=%u overlimits=%u bps=%u pps=%u qlen=%u backlog=%u",
		O->stats.bytes, O->stats.packets, O->stats.drops, O->stats.overlimits,
		O->stats.bps, O->stats.pps, O->stats.qlen, O->stats.backlog);

	return ret;
}

unsigned long lq_usec2tick(long usec)
{
	return usec * lq_tick_in_usec;
}

unsigned long lq_tick2usec(long tick)
{
	return tick / lq_tick_in_usec;
}

unsigned long lq_xmittime(unsigned long rate, unsigned long size)
{
	return lq_usec2tick(1000000 * ((double) size / rate));
}

int lq_read_psched(void)
{
	FILE	*f;

	f = fopen("/proc/net/psched", "r");
	if (!f)
		return -1;

	if (fscanf(f, "%08x%08x", &lq_t2us, &lq_us2t) != 2) {
		fclose (f);
		return -1;
	}

	fclose(f);

	lq_tick_in_usec = (double) lq_t2us / lq_us2t;

	return 0;
}

int lq_compute_rtable(unsigned long bps, __u32 *rtab, int cell_log, unsigned int mtu, unsigned int mpu)
{
	int		i;
	unsigned int	sz;

	if (mtu == 0)
		mtu = 1500;

	if (cell_log < 0) {
		cell_log = 0;
		while ((mtu >> cell_log) > 255)
			cell_log++;
	}

	for (i = 0; i < 256; i++) {
		sz = i << cell_log;
		if (sz < mpu)
			sz = mpu;
		rtab [i] = lq_usec2tick(1000000 * ((double) sz / bps));
	}

	return cell_log;
}

void lq_print_rate(char *buf, unsigned int len, unsigned long rate)
{
	snprintf(buf, len, "%lu bytes/s", rate);
}

// rate is in bits
void lq_rate2string(u_int64_t rate, char *str, int size)
{
	if (rate >= 1000000000)
		snprintf(str, size, "%lluGbit", rate / 1000000000);
	else if (rate >= 1000000)
		snprintf(str, size, "%lluMbit", rate / 1000000);
	else if (rate >= 1000)
		snprintf(str, size, "%lluKbit", rate / 1000);
	else
		snprintf(str, size, "%llubit", rate);
}

