#include <lq_com.h>
#include <lq_nl_com.h>


/* Types */

/* Global variables */
short		lq_nl_debug = 0;


/* Internal variables */
static int		lq_nl_fd;
static short		lq_nl_inited = 0;
static char		lq_nl_strerror0[256];
static pid_t		lq_nl_pid;
static u_int32_t	lq_nl_seq;
/* hash for interfaces */
struct lq_nl_if {
	char			name[64];
	unsigned char		family;
	unsigned short		type;
	int			ifindex;
	unsigned int		flags;
	int			addr;
	unsigned int		mtu;
	struct lq_nl_if		*next;
	/*struct rtnl_link_stats	stats;*/
};

static struct lq_nl_if		lq_nl_ifs[16];
void				(*lq_nl_send_cb)(struct nlmsghdr *h) = NULL;
static void			*lq_nl_send_cb_arg = NULL;

/* Prototypes */
struct lq_nl_if		*lq_nl_getif_byindex(int index);


/* Functions */
char *lq_nl_strerror(void)
{
	return lq_nl_strerror0;
}

char *lq_nl_getnltype(int type)
{
	char		*ret;

	switch (type) {
		case NLMSG_DONE:	ret = "NLMSG_DONE";	break;
		case NLMSG_ERROR:	ret = "NLMSG_ERROR";	break;

		case RTM_NEWLINK:	ret = "RTM_NEWLINK";	break;
		case RTM_DELLINK:	ret = "RTM_DELLINK";	break;
		case RTM_GETLINK:	ret = "RTM_GETLINK";	break;

		case RTM_NEWADDR:	ret = "RTM_NEWADDR";	break;
		case RTM_DELADDR:	ret = "RTM_DELADDR";	break;
		case RTM_GETADDR:	ret = "RTM_GETADDR";	break;

		case RTM_NEWROUTE:	ret = "RTM_NEWROUTE";	break;
		case RTM_DELROUTE:	ret = "RTM_DELROUTE";	break;
		case RTM_GETROUTE:	ret = "RTM_GETROUTE";	break;

		case RTM_NEWNEIGH:	ret = "RTM_NEWNEIGH";	break;
		case RTM_DELNEIGH:	ret = "RTM_DELNEIGH";	break;
		case RTM_GETNEIGH:	ret = "RTM_GETNEIGH";	break;

		case RTM_NEWRULE:	ret = "RTM_NEWRULE";	break;
		case RTM_DELRULE:	ret = "RTM_DELRULE";	break;
		case RTM_GETRULE:	ret = "RTM_GETRULE";	break;

		case RTM_NEWQDISC:	ret = "RTM_NEWQDISC";	break;
		case RTM_DELQDISC:	ret = "RTM_DELQDISC";	break;
		case RTM_GETQDISC:	ret = "RTM_GETQDISC";	break;

		case RTM_NEWTCLASS:	ret = "RTM_NEWTCLASS";	break;
		case RTM_DELTCLASS:	ret = "RTM_DELTCLASS";	break;
		case RTM_GETTCLASS:	ret = "RTM_GETTCLASS";	break;

		case RTM_NEWTFILTER:	ret = "RTM_NEWTFILTER";	break;
		case RTM_DELTFILTER:	ret = "RTM_DELTFILTER";	break;
		case RTM_GETTFILTER:	ret = "RTM_GETFILTER";	break;

		default:		ret = "UNKNOWN!";
	}

	return ret;
}

static void lq_nl_parse_rtattr(struct rtattr *tb[], int max, struct rtattr *rta, int len)
{
	memset(tb, 0, sizeof(struct rtattr) * (max + 1));
	while (RTA_OK(rta, len)) {
		if (rta->rta_type <= max)
			tb[rta->rta_type] = rta;
		rta = RTA_NEXT(rta, len);
	}

	if (len)
		fprintf(stderr, "Broken message [len=%d]!\n",
			len);
}

static void lq_nl_parse_rtattr_nested(struct rtattr *tb[], int max, struct rtattr *rta)
{
	lq_nl_parse_rtattr(tb, max, RTA_DATA(rta), RTA_PAYLOAD(rta));
}

char *lq_nl_dump_nlflags(__u16 flags)
{
	static char	ret[128];
	char		add[4];

	strcpy(ret, "");
	strcpy(add, "");

	if (flags & NLM_F_REQUEST) {
		strcat(ret, add);
		strcat(ret, "NLM_F_REQUEST");
		strcpy(add, " | ");
	}
	if (flags & NLM_F_MULTI) {
		strcat(ret, add);
		strcat(ret, "NLM_F_MULTI");
		strcpy(add, " | ");
	}
	if (flags & NLM_F_ACK) {
		strcat(ret, add);
		strcat(ret, "NLM_F_ACK");
		strcpy(add, " | ");
	}
	if (flags & NLM_F_ECHO) {
		strcat(ret, add);
		strcat(ret, "NLM_F_ECHO");
		strcpy(add, " | ");
	}

	return ret;
}

char *lq_nl_dump_nlmsghdr(struct nlmsghdr *hdr)
{
	static char	ret[256];
	char		*type;
	char		*flags;

	type = lq_nl_getnltype(hdr->nlmsg_type);
	flags = lq_nl_dump_nlflags(hdr->nlmsg_flags);

	snprintf(ret, sizeof(ret),
		"len=%u type=%u (%s) flags=0x%x (%s) seq=%u pid=%u",
		hdr->nlmsg_len, hdr->nlmsg_type, type,
		hdr->nlmsg_flags, flags,
		hdr->nlmsg_seq, hdr->nlmsg_pid);

	return ret;
}

char *lq_nl_dump_tcmsg(struct tcmsg *tc)
{
	static char	ret[128];

	snprintf(ret, sizeof(ret),
		"family=%u ifindex=%d handle=0x%x parent=0x%x info=0x%x",
		tc->tcm_family, tc->tcm_ifindex, tc->tcm_handle,
		tc->tcm_parent, tc->tcm_info);

	return ret;
}

char *lq_nl_dump(void *data)
{
	static char		ret[40960];
	char			tmp[256];
	unsigned int		i, rest, hdrlen, glen;
	unsigned char		*p;
	struct lq_nl_pak	*req;
	struct rtattr		*rta;
	struct nlmsghdr		*n;

	strcpy(ret, "");

	if (!data) {
		strcat (ret, "Invalid pointer!");
		goto out;
	}

	req = (struct lq_nl_pak *) data;
	glen = req->n.nlmsg_len;
	n = &req->n;

	strcat(ret, "\tRAW data: ");
	p = (unsigned char *) data;
	for (i = 0; i < glen; i ++) {
		if (i % 16 == 0)
			strcat(ret, "\n\t");
		snprintf(tmp, sizeof(tmp), "%02x ", p[i]);
		strcat(ret, tmp);
	}
	strcat(ret, "\n");

	strcat(ret, "\tnlmsghdr: ");
	strcat(ret, lq_nl_dump_nlmsghdr(n));
	strcat(ret, "\n");

	if (req->n.nlmsg_type == NLMSG_DONE) {
		strcat(ret, "\tNLMSG_DONE\n");
	} else if (req->n.nlmsg_type == NLMSG_ERROR) {
		struct nlmsgerr	*err;

		err = (struct nlmsgerr *) NLMSG_DATA(n);
		snprintf(tmp, sizeof(tmp), "\tNLMSG_ERROR: code=%d [%s]",
			err->error, strerror(-err->error));
		strcat(ret, tmp);
		goto out;
	} else if ((n->nlmsg_type >= RTM_NEWQDISC) && (n->nlmsg_type <= RTM_GETTFILTER)) {
		strcat(ret, "\ttcmsg: ");
		strcat(ret, lq_nl_dump_tcmsg(&req->tc));
		strcat(ret, "\n");

		p = (unsigned char *) &req->buf;
		hdrlen = (char *) &req->buf - (char *) &req->n;
		rest = req->n.nlmsg_len - hdrlen;

		rta = (struct rtattr *) p;
		while (RTA_OK(rta, rest)) {
			unsigned char	*pl;
			int		pl_len;

			pl = RTA_DATA(rta);
			pl_len = RTA_PAYLOAD(rta);

			snprintf(tmp, sizeof(tmp), "\t[len=%2d] [type=%2d] [payload @ %p] [payload len %2d]",
				rta->rta_len, rta->rta_type,
				pl, pl_len);
			strcat(ret, tmp);

			for (i = 0; i < pl_len; i ++) {
				if (i % 16 == 0)
					strcat(ret, "\n\t");
				snprintf(tmp, sizeof(tmp), "%02x ", pl[i]);
				strcat(ret, tmp);
			}
			strcat(ret, "\n");

			rta = RTA_NEXT(rta, rest);
		}
	} else if ((n->nlmsg_type >= RTM_NEWLINK) && (n->nlmsg_type <= RTM_GETLINK)) {
		struct ifinfomsg	*if0;

		if0 = (struct ifinfomsg *) RTA_DATA(n);
		snprintf(tmp, sizeof(tmp), "\tifinfomsg: family=%d type=%d index=%d flags=0x%x change=%d",
			if0->ifi_family, if0->ifi_type,
			if0->ifi_index, if0->ifi_flags,
			if0->ifi_change);
		strcat(ret, tmp);
	} else {
		snprintf(tmp, sizeof(tmp),
			"DEBUG: Unknown type [%d 0x%x]",
			n->nlmsg_type, n->nlmsg_type);
		strcat(ret, tmp);
	}

	out:
	return ret;
}

char *lq_nl_dump_msg(struct msghdr *hdr)
{
	static char		ret[40960];
	char			*nl;
	struct lq_nl_pak	*pak;
	struct iovec		*iov;

	snprintf(ret, sizeof(ret),
		"name=%p namelen=%u iov=%p iovlen=%u control=%p controllen=%u flags=0x%x\n",
		hdr->msg_name, hdr->msg_namelen,
		(void *) hdr->msg_iov, hdr->msg_iovlen,
		hdr->msg_control, hdr->msg_controllen,
		hdr->msg_flags);

	iov = (struct iovec *) hdr->msg_iov;
	pak = (struct lq_nl_pak *) iov->iov_base;
	nl = lq_nl_dump(pak);
	strcat(ret, nl);

	return ret;
}

int lq_nl_recvmsg(struct msghdr *msg)
{
	int	ret = 0, n;

	while (1) {
		n = recvmsg(lq_nl_fd, msg, 0);
		if ((n < 0) && (errno == EINTR))
			continue;

		if (n <= 0) {
			snprintf(lq_nl_strerror0, sizeof(lq_nl_strerror0),
				"nl_recvmsg: error [%s]",
				strerror(errno));
			ret = -1;
			goto out;
		}

		ret = n;
		break;
	} /* while */

	out:
	if (lq_nl_debug > 4)
		printf("nl_recvmsg: Dump received message [n=%d]:\n\t%s\n",
			n, lq_nl_dump_msg(msg));

	return ret;
}

int lq_nl_sendmsg(struct msghdr *msg)
{
	int	ret = 0, n;

	while (1) {
		n = sendmsg(lq_nl_fd, msg, 0);
		if ((n < 0) && (errno == EINTR))
				continue;

		if (n <= 0) {
			snprintf(lq_nl_strerror0, sizeof(lq_nl_strerror0),
				"nl_sendmsg: error [%s]",
				strerror(errno));
			ret = -1;
			goto out;
		}

		ret = n;
		break;
	} /* while */

	out:
	if (lq_nl_debug > 4)
		printf("nl_sendmsg: Dump sent message [n=%d]:\n\t%s\n",
			n, lq_nl_dump_msg(msg));

	return ret;
}

int lq_nl_send(struct lq_nl_pak *req)
{
	int			ret = 0;
	int			status;
	struct nlmsghdr		*n;
	struct nlmsghdr		*h;
	struct iovec		iov;
	char			buf[8192];
	struct msghdr		msg;
	struct sockaddr_nl	addr;

	if (lq_nl_debug > 1)
		printf("DEBUG: %s\n", __FUNCTION__);

	if (lq_nl_inited == 0) {
		snprintf(lq_nl_strerror0, sizeof(lq_nl_strerror0),
			"nl_send: init was not called first");
		ret = -1;
		goto out;
	}

	n = &req->n;

	iov.iov_base = (void *) n;
	iov.iov_len = n->nlmsg_len;

	msg.msg_name = (void *) &addr;
	msg.msg_namelen = sizeof(addr);
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;
	msg.msg_control = NULL;
	msg.msg_controllen = 0;
	msg.msg_flags = 0;

	memset((void *) &addr, 0, sizeof(addr));
	addr.nl_family = AF_NETLINK;
	addr.nl_pid = 0;
	addr.nl_groups = 0;
	n->nlmsg_seq = ++lq_nl_seq;
	n->nlmsg_pid = lq_nl_pid;

	ret = lq_nl_sendmsg(&msg);
	if (ret == -1)
		goto out;

	if (lq_nl_debug > 3)
		printf("DEBUG: Start getting messages from nl...\n");

	while (1) {
		iov.iov_base = buf;
		iov.iov_len = sizeof(buf);

		status = lq_nl_recvmsg(&msg);
		if (status <= 0)
			goto out;

		h = (struct nlmsghdr *) buf;
		while (NLMSG_OK(h, status)) {
			if (lq_nl_debug > 0)
				printf("nl_send: Parse received data [status=%d]\n%s\n",
					status, lq_nl_dump(h));

			/* is for our pid? */
			if (h->nlmsg_pid != lq_nl_pid) {
				if (lq_nl_debug > 1)
					printf("Invalid pid [%d] [%d]! Continue.\n",
						h->nlmsg_pid, lq_nl_pid);
				h = NLMSG_NEXT(h, status);
				continue;
			}

			/* seq match? */
			if (h->nlmsg_seq != lq_nl_seq) {
				if (lq_nl_debug > 1)
					printf("Invalid seq [%u] [%u]! Continue.\n",
						h->nlmsg_seq, lq_nl_seq);
				h = NLMSG_NEXT(h, status);
				continue;
			}

			if (h->nlmsg_type == NLMSG_ERROR) {
				struct nlmsgerr	*err;

				err = (struct nlmsgerr *) NLMSG_DATA(h);
				if (err->error != 0) {
					snprintf(lq_nl_strerror0, sizeof(lq_nl_strerror0),
						"send: error code = %d", err->error);
					ret = -1;
				} else {
					ret = 0;
				}
				goto out;
			}

			if (h->nlmsg_type == NLMSG_DONE) {
				ret = 0;
				goto out;
			}

			if (lq_nl_send_cb)
				lq_nl_send_cb(h);

			h = NLMSG_NEXT(h, status);
		} /* while - every struct in message */
	} /* while */

	out:
	return ret;
}

void lq_nl_get_type_n_flags(short cmd, int type, unsigned int *mtype, unsigned int *flags)
{
	int	rf = 0;

	switch (cmd) {
		case LQ_add:
			rf = NLM_F_EXCL | NLM_F_CREATE;
			switch (type) {
				case LQ_qdisc:	*mtype = RTM_NEWQDISC;		break;
				case LQ_class:	*mtype = RTM_NEWTCLASS;		break;
				case LQ_filter:	*mtype = RTM_NEWTFILTER;	break;
			}
			break;

		case LQ_del:
			rf = 0;
			switch (type) {
				case LQ_qdisc:	*mtype = RTM_DELQDISC;		break;
				case LQ_class:	*mtype = RTM_DELTCLASS;		break;
				case LQ_filter:	*mtype = RTM_DELTFILTER;	break;
			}
			break;

		case LQ_change:
			rf = 0;
			switch (type) {
				case LQ_qdisc:	*mtype = RTM_NEWQDISC;		break;
				case LQ_class:	*mtype = RTM_NEWTCLASS;		break;
				case LQ_filter:	*mtype = RTM_NEWTFILTER;	break;
			}
			break;
	}

	*flags = rf;
}

/* append at the end of pak a rtattr structure */
/* in the end, update the len */
void lq_nl_attr(struct lq_nl_pak *req, int type, void *data, int datalen)
{
	int		oldlen;
	struct nlmsghdr	*n;
	struct rtattr	*rta;
	void		*p;

	if (lq_nl_debug > 7)
		printf("\tDEBUG: %s(req, type=%d, data, datalen=%d)\n",
			__FUNCTION__, type, datalen);

	n = &req->n;
	oldlen = NLMSG_ALIGN(n->nlmsg_len);
	printf("\tDEBUG: NLMSG_ALIGN(oldlen)=%d\n", oldlen);

	rta = (struct rtattr *)(((char *) n) + oldlen);
	rta->rta_type = type;
	rta->rta_len = RTA_LENGTH(datalen);
	printf("\tDEBUG: rta_len=%d\n", rta->rta_len);

	if (datalen > 0) {
		p = RTA_DATA(rta);
		memcpy(p, data, datalen);
	}
	n->nlmsg_len = oldlen + rta->rta_len;
	printf("\tDEBUG: new nlmsg_len = %d (+%d)\n",
		n->nlmsg_len, rta->rta_len);
}

int lq_nl_q_encode(lq *O, struct lq_nl_pak *req)
{
	int	ret = 0;

	switch (O->subtype) {
		case LQ_q_htb:
			lq_nl_attr(req, TCA_OPTIONS, NULL, 0);
			lq_nl_attr(req, TCA_HTB_INIT, &O->data.q.htb.glob, NLMSG_ALIGN(sizeof(O->data.q.htb.glob)));
			break;

		case LQ_q_sfq:
			lq_nl_attr(req, TCA_OPTIONS, &O->data.q.sfq.opt, sizeof(O->data.q.sfq.opt));
			break;

		case LQ_q_pfifo:
		case LQ_q_bfifo:
			lq_nl_attr(req, TCA_OPTIONS, &O->data.q.fifo.opt, sizeof(O->data.q.fifo.opt));
			break;

		case LQ_q_prio:
			lq_nl_attr(req, TCA_OPTIONS, &O->data.q.prio.opt, sizeof(O->data.q.prio.opt));
			break;

		default:
			; /* do nothing */
	}

	return ret;
}

int lq_nl_c_encode(lq *O, struct lq_nl_pak *req)
{
	int	ret = 0;

	switch (O->subtype) {
		case LQ_c_htb: {
			__u32 rtab[256], ctab[256];
			unsigned int mtu;
			struct lq_nl_if *if0;
			int cell_log;

			if0 = lq_nl_getif_byindex(O->ifindex);
			if (!if0) {
				snprintf(lq_nl_strerror0, sizeof(lq_nl_strerror0),
					"invalid interface");
				ret = -1;
				goto out;
			}

			mtu = if0->mtu;
			if (mtu == 0)
				mtu = 1500;

			if (O->data.c.htb.opt.rate.rate == 0) {
				snprintf(lq_nl_strerror0, sizeof(lq_nl_strerror0),
					"rate cannot be 0");
				ret = -1;
				goto out;
			}

			if (O->data.c.htb.opt.ceil.rate == 0)
				O->data.c.htb.opt.ceil.rate = O->data.c.htb.opt.ceil.rate;

			if (O->data.c.htb.opt.rate.rate > O->data.c.htb.opt.ceil.rate) {
				snprintf(lq_nl_strerror0, sizeof(lq_nl_strerror0),
					"rate is greater than ceil");
				ret = -1;
				goto out;
			}

			/* fix buffer and cbuffer */
			if (O->data.c.htb.opt.buffer == 0)
				O->data.c.htb.opt.buffer = O->data.c.htb.opt.rate.rate / HZ + mtu;

			if (O->data.c.htb.opt.cbuffer == 0)
				O->data.c.htb.opt.cbuffer = O->data.c.htb.opt.ceil.rate / HZ + mtu;

			cell_log = lq_compute_rtable(O->data.c.htb.opt.rate.rate, rtab, -1, mtu, 0);
			if (cell_log < 0) {
				snprintf(lq_nl_strerror0, sizeof(lq_nl_strerror0),
					"error computing rate table");
				ret = -1;
				goto out;
			}

			O->data.c.htb.opt.buffer = lq_xmittime(O->data.c.htb.opt.rate.rate, O->data.c.htb.opt.buffer);
			O->data.c.htb.opt.rate.cell_log = cell_log;

			cell_log = lq_compute_rtable(O->data.c.htb.opt.ceil.rate, ctab, -1, mtu, 0);
			if (cell_log < 0) {
				snprintf(lq_nl_strerror0, sizeof(lq_nl_strerror0),
					"error computing ceil rate table");
				ret = -1;
				goto out;
			}

			O->data.c.htb.opt.cbuffer = lq_xmittime(O->data.c.htb.opt.ceil.rate, O->data.c.htb.opt.cbuffer);
			O->data.c.htb.opt.ceil.cell_log = cell_log;

			lq_nl_attr(req, TCA_OPTIONS, NULL, 0);
			lq_nl_attr(req, TCA_HTB_PARMS, &O->data.c.htb.opt, NLMSG_ALIGN(sizeof(O->data.c.htb.opt)));
			lq_nl_attr(req, TCA_HTB_RTAB, rtab, sizeof(rtab));
			lq_nl_attr(req, TCA_HTB_CTAB, ctab, sizeof(ctab));
			} break;
	}

	out:
	return ret;
}

int lq_nl_f_encode(lq *O, struct lq_nl_pak *req)
{
	if (O->data.f.protocol == 0)
		O->data.f.protocol = htons(ETH_P_IP);

	req->tc.tcm_info = (O->data.f.prio << 16) | O->data.f.protocol;

	switch (O->subtype) {
		case LQ_f_u32:
			lq_nl_attr(req, TCA_OPTIONS, NULL, 0);
			break;

		case LQ_f_fw:
			lq_nl_attr(req, TCA_OPTIONS, NULL, 0);
			lq_nl_attr(req, TCA_FW_CLASSID, &O->data.f.classid, sizeof(O->data.f.classid));
			break;

		default:
			snprintf(lq_nl_strerror0, sizeof(lq_nl_strerror0),
				"Filter not known");
			return -1;
	}

	return 0;
}

int lq_nl_encode_type(lq *O, struct lq_nl_pak *req, char *type)
{
	lq_nl_attr(req, TCA_KIND, type, strlen(type) + 1);

	return 0;
}

int lq_nl_encode(short cmd, lq *O, struct lq_nl_pak *req)
{
	int		ret = -1;
	int		len;
	struct rtattr	*rta;
	unsigned int	mtype, flags;
	char		*stype;

	memset((void *) req, 0, sizeof(*req));

	lq_nl_get_type_n_flags(cmd, O->type, &mtype, &flags);

	req->n.nlmsg_len = NLMSG_LENGTH(sizeof(struct tcmsg));
	req->n.nlmsg_type = mtype;
	req->n.nlmsg_seq = ++lq_nl_seq;
	req->n.nlmsg_pid = lq_nl_pid;
	req->n.nlmsg_flags = NLM_F_REQUEST | NLM_F_ACK | flags;

	req->tc.tcm_family = AF_UNSPEC;
	req->tc.tcm_ifindex = O->ifindex;
	req->tc.tcm_handle = O->id;
	req->tc.tcm_parent = O->parent;

	if (cmd == LQ_add) {
		stype = lq_subtype(O);
		lq_nl_encode_type(O, req, stype);
	}

	/* save len now */
	len = NLMSG_ALIGN(req->n.nlmsg_len);
	printf("DEBUG: Start len is s%d\n", len);

	switch (O->type) {
		case LQ_qdisc:
			ret = lq_nl_q_encode(O, req);
			break;

		case LQ_class:
			ret = lq_nl_c_encode(O, req);
			break;

		case LQ_filter:
			ret = lq_nl_f_encode(O, req);
			break;

		default:
			snprintf(lq_nl_strerror0, sizeof(lq_nl_strerror0),
				"Invalid type %d", O->type);
			return -1;
	}

	/* modify len of TCA_OPTIONS */
	rta = (struct rtattr *)((char *) &req->n + len);
	printf("DEBUG: update rta_len. now is %d.\n", rta->rta_len);
	rta->rta_len = req->n.nlmsg_len - len;
	printf("DEBUG: update rta_len. NOW is %d.\n", rta->rta_len);

	return ret;
}

void lq_nl_getstats_func(struct nlmsghdr *h)
{
	struct tcmsg	*t = NLMSG_DATA(h);
	int		len = h->nlmsg_len;
	lq_stats	ls;
	void		(*f)(lq_stats *) = lq_nl_send_cb_arg;
	short		callme = 0;
	struct rtattr	*tb[TCA_MAX + 1];

	if (lq_nl_debug > 4)
		fprintf(stderr, "%s(%p)\n", __FUNCTION__, (void *) h);

	if (!f)
		return;

	len -= NLMSG_LENGTH(sizeof(*t));

	memset(tb, 0, sizeof(tb));
	lq_nl_parse_rtattr(tb, TCA_MAX, TCA_RTA(t), len);

	if (tb[TCA_KIND] == NULL) {
		fprintf(stderr, "Invalid KIND type!\n");
		return;
	}

	if (tb[TCA_STATS]) {
		struct tc_stats *s = RTA_DATA(tb[TCA_STATS]);

		/*fprintf(stderr, "TCA_STATS\n");*/
		callme = 1;

		ls.ifindex = t->tcm_ifindex;
		ls.handle = t->tcm_handle;
		ls.parent = t->tcm_parent;
		ls.bytes = s->bytes;
		ls.packets = s->packets;
		ls.drops = s->drops;
		ls.overlimits = s->overlimits;
		ls.bps = s->bps;
		ls.pps = s->pps;
		ls.qlen = s->qlen;
		ls.backlog = s->backlog;
	}

	if (tb[TCA_XSTATS]) {
		/*fprintf(stderr, "TCA_XSTATS\n");*/
	}

	if (tb[TCA_KIND]) {
		/*
		fprintf(stderr, "TCA_KIND=[%s]\n",
			(char *) RTA_DATA(tb[TCA_KIND]));
		*/
	}

	if (tb[TCA_OPTIONS]) {
		/*fprintf(stderr, "TCA_OPTIONS\n");*/

		if (!strncmp(RTA_DATA(tb[TCA_KIND]), "htb", 3)) {
			struct rtattr *tb_htb[TCA_HTB_RTAB+1];

			memset(tb_htb, 0, sizeof(tb_htb));
			lq_nl_parse_rtattr_nested(tb_htb, TCA_HTB_RTAB, tb[TCA_OPTIONS]);

			if (tb_htb[TCA_HTB_PARMS]) {
				struct tc_htb_opt *opt;

				callme = 1;

				opt = RTA_DATA(tb_htb[TCA_HTB_PARMS]);

				ls.rate = opt->rate.rate;
				ls.ceil = opt->ceil.rate;
			}
		}
	}

	if (callme == 1)
		f(&ls);
}

int lq_nl_getstats(short type, int ifindex, void(*f)(lq_stats *))
{
	int			ret;
	struct lq_nl_pak	req;

	memset(&req, 0, sizeof(req));
	req.n.nlmsg_len = NLMSG_LENGTH(sizeof(struct tcmsg));
	req.n.nlmsg_flags = NLM_F_ROOT | NLM_F_MATCH | NLM_F_REQUEST;
	switch (type) {
		case LQ_qdisc:
			req.n.nlmsg_type = RTM_GETQDISC;
			break;

		case LQ_class:
			req.n.nlmsg_type = RTM_GETTCLASS;
			break;

		case LQ_filter:
			req.n.nlmsg_type = RTM_GETTFILTER;
			break;

		default:
			snprintf(lq_nl_strerror0, sizeof(lq_nl_strerror0), 
				"Invalid type [%d]", type);
			return -1;
	}

	req.tc.tcm_family = AF_UNSPEC;
	req.tc.tcm_ifindex = ifindex;

	/* set callback */
	lq_nl_send_cb = lq_nl_getstats_func;
	lq_nl_send_cb_arg = (void *) f;

	ret = lq_nl_send(&req);
	if (ret == -1)
		return -1;

	return 0;
}


/*///////////////////////// Interface */
void lq_nl_addif(struct lq_nl_if *dev)
{
	int			index;
	struct lq_nl_if	*p;

	index = dev->ifindex & 0x0f;
	p = &lq_nl_ifs[index];
	if (p->ifindex > 0) {
		while (p->next)
			p = p->next;

		/* alloc structure */
		p->next = (struct lq_nl_if *) malloc(sizeof(struct lq_nl_if));
		if (!p->next)
			return;
		p = p->next;
	}

	memcpy(p, dev, sizeof(struct lq_nl_if));
	p->next = NULL;
}

struct lq_nl_if *lq_nl_getif_byindex(int index)
{
	int			i;
	struct lq_nl_if	*p;

	for(i = 0; i < 16; i++) {
		p = &lq_nl_ifs[i];
		while (p) {
			if (index == p->ifindex)
				goto out;

			p = p->next;
		}
	}

	out:
	return p;
}

int lq_nl_getif(char *dev)
{
	int			i;
	struct lq_nl_if	*p;
	int			ret = -1;

	for (i = 0; i < 16; i ++) {
		p = &lq_nl_ifs[i];
		while (p) {
			if (strcmp(p->name, dev) == 0) {
				ret = p->ifindex;
				goto out;
			}
			p = p->next;
		}
	}	

	snprintf(lq_nl_strerror0, sizeof(lq_nl_strerror0),
		"Cannot get info about interface.");

	out:
	return ret;
}

void lq_nl_listifs(void) {
	int			i;
	struct lq_nl_if	*p;

	for (i = 0; i < 16; i ++) {
		p = &lq_nl_ifs[i];
		while (p) {
			if (p->ifindex > 0)
				printf("index=%3d family=0x%4x type=0x%4x mtu=%6u flags=%4x name=%s\n",
					p->ifindex, p->family, p->type, p->mtu, p->flags, p->name);
			p = p->next;
		}
	}
}

int lq_nl_getifs(void)
{
	struct ifinfomsg		*ifi;
	int				nr;
	int				ret = -1;
	struct	msghdr			msg;
	char				buf[8192];
	struct iovec			iov;
	struct sockaddr_nl		addr;
	struct nlmsghdr			*n;
	struct lq_nl_generic_t	req;
	struct lq_nl_if		if0;
	struct rtattr *tb[IFLA_MAX + 1];
	int len;

	memset((void *) &addr, 0, sizeof(addr));
	addr.nl_family = AF_NETLINK;

	memset((void *) &req, 0, sizeof(req));
	req.n.nlmsg_len = sizeof(req);
	req.n.nlmsg_type = RTM_GETLINK;
	req.n.nlmsg_flags = NLM_F_ROOT | NLM_F_MATCH | NLM_F_REQUEST;
	req.n.nlmsg_pid = lq_nl_pid;
	lq_nl_seq ++;
	req.n.nlmsg_seq = lq_nl_seq;
	req.g.rtgen_family = AF_UNSPEC;

	iov.iov_base = (void *) &req;
	iov.iov_len = req.n.nlmsg_len;

	msg.msg_name = (void *) &addr;
	msg.msg_namelen = sizeof(addr);
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;
	msg.msg_control = NULL;
	msg.msg_controllen = 0;
	msg.msg_flags = 0;

	while (1) {
		nr = lq_nl_sendmsg(&msg);
		if (nr == -1) {
			snprintf(lq_nl_strerror0, sizeof(lq_nl_strerror0),
				"nl_getifs: cannot send [%s]", strerror(errno));
			ret = -1;
			goto out;
		}

		break;
	}

	/* get response */
	ret = -1;
	while (1) {
		iov.iov_base = buf;
		iov.iov_len = sizeof(buf);

		nr = lq_nl_recvmsg(&msg);
		if (nr == -1)
			goto out;

		/* for every interface... */
		n = (struct nlmsghdr *) buf;
		while (NLMSG_OK(n, nr)) {
			if (lq_nl_debug > 9)
				printf("Test if [%s]...\n",
					lq_nl_dump_nlmsghdr(n));

			if (n->nlmsg_pid != lq_nl_pid) {
				if (lq_nl_debug > 0)
					printf("nl_getifs: invalid pid [%d] [%d]!\n",
						n->nlmsg_pid, lq_nl_pid);
				n = NLMSG_NEXT(n, nr);
				continue;
			}

			if (n->nlmsg_seq != lq_nl_seq) {
				if (lq_nl_debug > 0)
					printf("nl_getifs: invalid seq [%d] [%d]!\n",
						n->nlmsg_seq, lq_nl_seq);
				n = NLMSG_NEXT(n, nr);
				continue;
			}

			if (n->nlmsg_type == NLMSG_DONE) {
				if (lq_nl_debug > 9)
					printf("nl_getif: NLMSG_DONE!\n");
				ret = 0;
				goto out;
			}

			if (n->nlmsg_type == NLMSG_ERROR) {
				struct nlmsgerr	*err;

				err = (struct nlmsgerr *) NLMSG_DATA(n);
				snprintf(lq_nl_strerror0, sizeof(lq_nl_strerror0),
					"if_get: nl parse error [%d]!",
					err->error);
				ret = -1;
				goto out;
			}

			if (lq_nl_debug > 9)
				printf("DEBUG: nlmsg_type - 0x%x\n",
					n->nlmsg_type);

			if (lq_nl_debug > 9)
				printf("nl_getif: for every rtattr\n");
			/* for every rtattr... */
			ifi = NLMSG_DATA(n);
			len = n->nlmsg_len - NLMSG_LENGTH(sizeof(*ifi));
			lq_nl_parse_rtattr(tb, IFLA_MAX, IFLA_RTA(ifi), len);
			if (tb[IFLA_IFNAME] == NULL)
				continue;

			memset(&if0, 0, sizeof(if0));
			if0.ifindex = ifi->ifi_index;
			if0.flags = ifi->ifi_flags;
			if0.family = ifi->ifi_family;
			if0.type = ifi->ifi_type;

			snprintf(if0.name, sizeof(if0.name), "%s", (char *) RTA_DATA(tb[IFLA_IFNAME]));
			printf("name=%s\n", if0.name);

			if (tb[IFLA_MTU]) {
				if0.mtu = *(unsigned int *) RTA_DATA(tb[IFLA_MTU]);
				printf("mtu:%u\n", if0.mtu);
			}

			/*
			if (tb[IFLA_STATS]) {
				memcpy(&if0.stats, RTA_DATA(tb[IFLA_STATS]), sizeof(if0.stats));
				printf("stats: rx_bytes = %u\n", if0.stats.rx_bytes);
			}
			*/

			lq_nl_addif(&if0);
			printf("\n");

			n = NLMSG_NEXT(n, nr);
		} /* while */

		if (msg.msg_flags & MSG_TRUNC) {
			if (lq_nl_debug > 0)
				printf("MSG_TRUNC!\n");
			continue;
		}
	} /* while */

	ret = 0;

	out:
	return ret;
}



/*///////////////////////////////////////////////////////////// Main ///// */
int lq_nl_init(void)
{
	int			ret = 0;
	struct sockaddr_nl	addr;
	int			i;

	if (lq_nl_inited == 1) {
		ret = 0;
		goto out;
	}

	lq_nl_pid = getpid();
	lq_nl_seq = time(NULL);

	lq_nl_fd = socket(AF_NETLINK, SOCK_RAW, NETLINK_ROUTE);
	if (lq_nl_fd == -1) {
		snprintf(lq_nl_strerror0, sizeof(lq_nl_strerror0),
			"nl_init: cannot open socket [%s]",
			strerror(errno));
		ret = -1;
		goto out;
		}

	memset((void *) &addr, 0, sizeof(addr));
	addr.nl_family = AF_NETLINK;
	addr.nl_groups = 0;

	ret = bind(lq_nl_fd, (struct sockaddr *) &addr, sizeof(addr));
	if (ret < 0) {
		snprintf(lq_nl_strerror0, sizeof(lq_nl_strerror0),
			"nl_init: cannot bind [%s]",
			strerror(errno));
		ret = -1;
		goto out_close;
	}

	/* interface hash */
	for (i = 0; i < 16; i ++)
		memset((void *) &lq_nl_ifs[i], 0, sizeof(struct lq_nl_if));
	ret = lq_nl_getifs();
	if (ret == -1) {
		snprintf(lq_nl_strerror0, sizeof(lq_nl_strerror0),
			"nl_init: cannot load interfaces");
		ret = -1;
		goto out_close;
	}

	lq_nl_inited = 1;
	goto out;

	out_close:
		close(lq_nl_fd);

	out:
	return ret;
}

int lq_nl_stop(void)
{
	int	ret = 0;

	if (lq_nl_inited == 0) {
		snprintf(lq_nl_strerror0, sizeof(lq_nl_strerror0),
			"nl_stop: init was not called first");
		ret = -1;
		goto out;
	}

	close(lq_nl_fd);

	lq_nl_inited = 0;

	out:
	return ret;
}


